import pytest

from huscy.attributes import services
from huscy.attributes.models import AttributeSchema

pytestmark = pytest.mark.django_db


def test_create_new_schema_if_none_exists(mocker):
    spy = mocker.spy(services, '_create_initial_attribute_schema')

    assert not AttributeSchema.objects.exists()

    services.get_attribute_schema()

    assert 1 == len(AttributeSchema.objects.all())
    spy.assert_called_once()


def test_get_latest_schema(attribute_schema_v4):
    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'string'},
            'attribute3': {'type': 'number'},
        },
    }

    assert expected == services.get_attribute_schema().schema


def test_get_schema_by_version(django_db_reset_sequences, attribute_schema_v4):
    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
        },
    }

    assert expected == services.get_attribute_schema(version=2).schema



from django import forms
from django.contrib import admin

from prettyjson import PrettyJSONWidget

from huscy.attributes import models


class AttributeSchemaAdminForm(forms.ModelForm):
    class Meta:
        model = models.AttributeSchema
        fields = '__all__'
        widgets = {
            'schema': PrettyJSONWidget(),
        }


class AttributeSchemaAdmin(admin.ModelAdmin):
    form = AttributeSchemaAdminForm

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


admin.site.register(models.AttributeSchema, AttributeSchemaAdmin)



from rest_framework import generics, mixins, permissions, viewsets
from rest_framework.response import Response

from huscy.attributes import models, serializer, services
from huscy.subjects.models import Subject


class AttributeSchemaView(generics.ListCreateAPIView):
    permission_classes = (permissions.DjangoModelPermissions, )
    queryset = models.AttributeSchema.objects.all()
    serializer_class = serializer.AttributeSchemaSerializer

    def list(self, request):
        attribute_schema = services.get_attribute_schema()
        serializer = self.get_serializer(attribute_schema)
        return Response(serializer.data)


class AttributeSetViewSet(mixins.RetrieveModelMixin, mixins.UpdateModelMixin,
                          viewsets.GenericViewSet):
    permission_classes = (permissions.IsAuthenticated, )
    queryset = Subject.objects.all()
    serializer_class = serializer.AttributeSetSerializer

    def get_object(self):
        subject = super().get_object()
        return services.get_or_create_attribute_set(subject)



from django.contrib.auth.models import Permission
from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_403_FORBIDDEN


def test_admin_user_can_create_attribute_schema(admin_client, attribute_schema_v2):
    response = create_attribute_schema(admin_client)

    assert HTTP_201_CREATED == response.status_code


def test_user_with_permission_can_create_attribute_schema(client, user, attribute_schema_v2):
    create_permission = Permission.objects.get(codename='add_attributeschema')
    user.user_permissions.add(create_permission)

    response = create_attribute_schema(client)

    assert HTTP_201_CREATED == response.status_code


def test_user_without_permission_cannot_create_attribute_schema(client, attribute_schema_v2):
    response = create_attribute_schema(client)

    assert HTTP_403_FORBIDDEN == response.status_code


def test_anonymous_user_can_create_attribute_schema(anonymous_client, attribute_schema_v2):
    response = create_attribute_schema(anonymous_client)

    assert HTTP_403_FORBIDDEN == response.status_code


def create_attribute_schema(client):
    return client.post(reverse('attributeschema-detail'), data=dict(schema={}), format='json')



from rest_framework import serializers

from huscy.attributes import models
from huscy.attributes.services import update_attribute_set


class AttributeSchemaSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.AttributeSchema
        fields = (
            'created_at',
            'id',
            'schema',
        )
        read_only_fields = 'created_at',

    def update(self, attribute_schema, validated_data):
        return models.AttributeSchema.objects.create(**validated_data)


class AttributeSetSerializer(serializers.ModelSerializer):
    attribute_schema = AttributeSchemaSerializer(read_only=True)
    attributes = serializers.JSONField()

    class Meta:
        model = models.AttributeSet
        fields = (
            'attribute_schema',
            'attributes',
        )

    def update(self, attribute_set, validated_data):
        return update_attribute_set(attribute_set, **validated_data)



from django.contrib.contenttypes.models import ContentType
from jsonschema.validators import Draft7Validator

from huscy.attributes.models import AttributeSchema, AttributeSet
from huscy.pseudonyms.services import get_or_create_pseudonym


def get_or_create_attribute_set(subject):
    content_type = ContentType.objects.get_by_natural_key('attributes', 'attributeset')
    pseudonym = get_or_create_pseudonym(subject, content_type)

    attribute_set, created = AttributeSet.objects.get_or_create(pseudonym=pseudonym.code)
    return attribute_set


def update_attribute_set(attribute_set, attributes, attribute_schema_version=None):
    if attribute_schema_version is None:
        attribute_schema = attribute_set.attribute_schema
    else:
        if attribute_schema_version < attribute_set.attribute_schema.pk:
            raise Exception('New version for attribute schema must be greater than or equals with '
                            'current attribute schema version.')
        attribute_schema = AttributeSchema.objects.get(pk=attribute_schema_version)

    Draft7Validator(attribute_schema.schema).validate(attributes)

    attribute_set.attributes = attributes
    attribute_set.attribute_schema = attribute_schema
    attribute_set.save()

    return attribute_set


def add_choices_property_to_attribute_schema(name, choices, parent=None, required=None):
    return add_property_to_attribute_schema(name, 'string', parent, required, enum=choices)


def add_integer_property_to_attribute_schema(name, minimum=None, maximum=None, parent=None,
                                             required=False):
    kwargs = {}
    if minimum is not None:
        kwargs['minimum'] = minimum
    if maximum is not None:
        kwargs['maximum'] = maximum
    return add_property_to_attribute_schema(name, 'integer', parent, required, **kwargs)


def add_property_to_attribute_schema(property_name, property_type='object', parent=None,
                                     required=False, **kwargs):
    attribute_schema = get_attribute_schema().schema

    if parent is None:
        _add_to_properties(attribute_schema, property_name, property_type, required, **kwargs)
    else:
        _add_to_parent(attribute_schema, property_name, property_type, parent, required, **kwargs)

    return AttributeSchema.objects.create(schema=attribute_schema)


def _add_to_properties(json_schema, property_name, property_type, required, **kwargs):
    json_schema['properties'][property_name] = {
        'type': property_type,
    }

    for key, value in kwargs.items():
        json_schema['properties'][property_name][key] = value

    if required:
        json_schema.setdefault('required', [])
        json_schema['required'].append(property_name)


def _add_to_parent(json_schema, property_name, property_type, parent, required, **kwargs):
    if parent not in json_schema['properties']:
        raise Exception(f'This parent property does not exist: ({parent}).')

    json_schema_parent = json_schema['properties'][parent]

    if not 'object' == json_schema_parent['type']:
        raise Exception('Parent property must be of type "object". '
                        f'Current type is "{json_schema_parent["type"]}".')

    json_schema_parent.setdefault('properties', {})
    _add_to_properties(json_schema_parent, property_name, property_type, required, **kwargs)


def get_attribute_schema(version=None):
    queryset = AttributeSchema.objects

    try:
        if version is None:
            return queryset.latest('pk')
        else:
            return queryset.get(pk=version)
    except AttributeSchema.DoesNotExist:
        return _create_initial_attribute_schema()


def _create_initial_attribute_schema():
    if AttributeSchema.objects.exists():
        raise Exception('Initial attribute schema already exist!')

    return AttributeSchema.objects.create(schema={
        'type': 'object',
        'properties': {},
    })



import pytest

from huscy.attributes.services import add_integer_property_to_attribute_schema, get_attribute_schema

pytestmark = pytest.mark.django_db


def test_add_integer_property_to_attribute_schema(attribute_schema_v2):
    add_integer_property_to_attribute_schema('attribute2')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'integer'},
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_integer_property_with_minimum_value(attribute_schema_v2):
    add_integer_property_to_attribute_schema('attribute2', minimum=15)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {
                'type': 'integer',
                'minimum': 15,
            },
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_integer_property_with_maximum_value(attribute_schema_v2):
    add_integer_property_to_attribute_schema('attribute2', maximum=15)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {
                'type': 'integer',
                'maximum': 15,
            },
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_integer_property_with_minimum_and_maximum_value(attribute_schema_v2):
    add_integer_property_to_attribute_schema('attribute2', minimum=5, maximum=15)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {
                'type': 'integer',
                'minimum': 5,
                'maximum': 15,
            },
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_integer_property_as_required(attribute_schema_v2):
    add_integer_property_to_attribute_schema('attribute2', required=True)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'integer'},
        },
        'required': ['attribute2'],
    }

    assert expected == get_attribute_schema().schema


def test_add_integer_property_to_parent(attribute_schema_v2):
    add_integer_property_to_attribute_schema('attribute2', parent='attribute1')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {
                'type': 'object',
                'properties': {
                    'attribute2': {'type': 'integer'},
                },
            },
        },
    }

    assert expected == get_attribute_schema().schema



import pytest

from jsonschema.exceptions import ValidationError

from huscy.attributes.services import add_choices_property_to_attribute_schema, update_attribute_set

pytestmark = pytest.mark.django_db


def test_update_attribute_set(attribute_set):
    attributes = attribute_set.attributes
    attributes['attribute3'] = 100.0

    update_attribute_set(attribute_set, attributes)

    expected = {
        'attribute1': {},
        'attribute2': 'any string',
        'attribute3': 100.0,
    }

    attribute_set.refresh_from_db()

    assert expected == attribute_set.attributes


def test_update_attribute_set_with_invalid_data(attribute_set):
    attributes = attribute_set.attributes
    attributes['attribute3'] = 'invalid'

    with pytest.raises(ValidationError):
        update_attribute_set(attribute_set, attributes)


def test_update_attribute_set_data_together_with_attribute_schema_version(attribute_set):
    latest_attribute_schema = add_choices_property_to_attribute_schema('attribute4', ['a', 'b'])

    attributes = attribute_set.attributes
    attributes['attribute4'] = 'a'

    update_attribute_set(attribute_set, attributes, latest_attribute_schema.pk)

    attribute_set.refresh_from_db()

    expected = {
        'attribute1': {},
        'attribute2': 'any string',
        'attribute3': 4.5,
        'attribute4': 'a',
    }

    assert attribute_set.attribute_schema == latest_attribute_schema
    assert expected == attribute_set.attributes


def test_update_attribute_schema_to_lower_version(attribute_set):
    with pytest.raises(Exception) as e:
        update_attribute_set(attribute_set, attribute_set.attributes,
                             attribute_set.attribute_schema.pk - 1)

    expected = ('New version for attribute schema must be greater than or equals with current '
                'attribute schema version.')

    assert expected == str(e.value)



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN


def test_admin_user_can_update_attribute_set(admin_client, attribute_set, subject):
    response = update_attribute_set(admin_client, attribute_set, subject)

    assert HTTP_200_OK == response.status_code


def test_user_without_permission_can_update_attribute_set(client, attribute_set, subject):
    response = update_attribute_set(client, attribute_set, subject)

    assert HTTP_200_OK == response.status_code


def test_anonymous_user_can_update_attribute_set(anonymous_client, attribute_set, subject):
    response = update_attribute_set(anonymous_client, attribute_set, subject)

    assert HTTP_403_FORBIDDEN == response.status_code


def update_attribute_set(client, attribute_set, subject):
    return client.put(
        reverse('attributeset-detail', kwargs=dict(pk=subject.id)),
        data={
            'attribute_schema_version': attribute_set.attribute_schema.pk,
            'attributes': {
                'attribute1': {},
                'attribute2': 'another string',
                'attribute3': 1.0
            },
        },
        format='json',
    )



import pytest

from huscy.attributes.services import add_choices_property_to_attribute_schema, get_attribute_schema

pytestmark = pytest.mark.django_db


def test_add_choices_property_to_attribute_schema(attribute_schema_v2):
    add_choices_property_to_attribute_schema('attribute2', choices=['a', 'b', 'c'])

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {
                'type': 'string',
                'enum': ['a', 'b', 'c'],
            },
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_choices_property_as_required(attribute_schema_v2):
    add_choices_property_to_attribute_schema('attribute2', choices=['a', 'b', 'c'], required=True)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {
                'type': 'string',
                'enum': ['a', 'b', 'c'],
            },
        },
        'required': ['attribute2'],
    }

    assert expected == get_attribute_schema().schema


def test_add_choices_property_to_parent(attribute_schema_v2):
    add_choices_property_to_attribute_schema('attribute2', choices=['a', 'b', 'c'],
                                             parent='attribute1')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {
                'type': 'object',
                'properties': {
                    'attribute2': {
                        'type': 'string',
                        'enum': ['a', 'b', 'c'],
                    },
                },
            },
        },
    }

    assert expected == get_attribute_schema().schema



from model_bakery import baker

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

from huscy.attributes.serializer import AttributeSchemaSerializer


def test_admin_user_can_retrieve_attribute_set(admin_client, attribute_set, subject):
    response = retrieve_attribute_set(admin_client, subject)

    assert HTTP_200_OK == response.status_code


def test_user_without_permission_can_retrieve_attribute_set(client, attribute_set, subject):
    response = retrieve_attribute_set(client, subject)

    assert HTTP_200_OK == response.status_code


def test_anonymous_user_cannot_retrieve_attribute_set(anonymous_client, attribute_set, subject):
    response = retrieve_attribute_set(anonymous_client, subject)

    assert HTTP_403_FORBIDDEN == response.status_code


def test_create_attribute_set_if_it_does_not_exist(django_db_reset_sequences, client,
                                                   attribute_schema_v2):
    subject = baker.make('subjects.Subject')

    response = retrieve_attribute_set(client, subject)

    expected = {
        'attributes': {},
        'attribute_schema': AttributeSchemaSerializer(attribute_schema_v2).data
    }

    assert HTTP_200_OK == response.status_code
    assert expected == response.json()


def retrieve_attribute_set(client, subject):
    return client.get(reverse('attributeset-detail', kwargs=dict(pk=subject.id)))



from datetime import datetime

import pytest

from huscy.attributes.models import AttributeSchema
from huscy.attributes.services import _create_initial_attribute_schema

pytestmark = pytest.mark.django_db


@pytest.mark.freeze_time('2000-01-01T10:00:00')
def test_create_initial_attribute_schema():
    assert not AttributeSchema.objects.exists()

    _create_initial_attribute_schema()

    attribute_schema = AttributeSchema.objects.get()
    assert attribute_schema.created_at == datetime(2000, 1, 1, 10)

    expected = {
        'type': 'object',
        'properties': {},
    }
    assert expected == attribute_schema.schema


def test_create_initial_attribute_schema_raises_error_if_one_exists():
    _create_initial_attribute_schema()

    with pytest.raises(Exception) as e:
        _create_initial_attribute_schema()

    assert 'Initial attribute schema already exist!' == str(e.value)



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN


def test_admin_user_can_retrieve_attribute_schema(admin_client, attribute_schema_v4):
    response = retrieve_attribute_schema(admin_client)

    assert HTTP_200_OK == response.status_code


def test_user_without_permission_can_retrieve_attribute_schema(client, attribute_schema_v4):
    response = retrieve_attribute_schema(client)

    assert HTTP_200_OK == response.status_code


def test_anonymous_user_cannot_retrieve_attribute_schema(anonymous_client, attribute_schema_v4):
    response = retrieve_attribute_schema(anonymous_client)

    assert HTTP_403_FORBIDDEN == response.status_code


def retrieve_attribute_schema(client):
    return client.get(reverse('attributeschema-detail'))



import pytest

from huscy.attributes.services import add_property_to_attribute_schema, get_attribute_schema

pytestmark = pytest.mark.django_db


def test_add_property_to_attribute_schema(attribute_schema_v2):
    add_property_to_attribute_schema('attribute2')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'object'},
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_property_with_type_string(attribute_schema_v2):
    add_property_to_attribute_schema('attribute2', 'string')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'string'},
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_required_property(attribute_schema_v2):
    add_property_to_attribute_schema('attribute2', 'string', required=True)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'string'},
        },
        'required': ['attribute2'],
    }

    assert expected == get_attribute_schema().schema


def test_add_property_to_parent(attribute_schema_v2):
    add_property_to_attribute_schema('attribute2', 'string', parent='attribute1')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {
                'type': 'object',
                'properties': {
                    'attribute2': {'type': 'string'},
                },
            },
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_required_property_to_parent(attribute_schema_v2):
    add_property_to_attribute_schema('attribute2', 'string', parent='attribute1', required=True)

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {
                'type': 'object',
                'properties': {
                    'attribute2': {'type': 'string'},
                },
                'required': ['attribute2'],
            },
        },
    }

    assert expected == get_attribute_schema().schema


def test_add_property_to_not_existing_parent(attribute_schema_v2):
    with pytest.raises(Exception) as e:
        add_property_to_attribute_schema('attribute2', 'string', parent='not_existing_property')

    assert 'This parent property does not exist: (not_existing_property).' == str(e.value)


def test_add_property_parent_that_is_not_type_object(attribute_schema_v3):
    with pytest.raises(Exception) as e:
        add_property_to_attribute_schema('attribute3', 'string', parent='attribute2')

    assert 'Parent property must be of type "object". Current type is "string".' == str(e.value)


def test_add_property_with_further_keyword_arguments(attribute_schema_v2):
    add_property_to_attribute_schema('attribute2', 'string', parent=None, required=False,
                                     extra1='extra1', extra2='extra2')

    expected = {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {
                'type': 'string',
                'extra1': 'extra1',
                'extra2': 'extra2',
            },
        },
    }

    assert expected == get_attribute_schema().schema



from setuptools import find_namespace_packages, setup

from huscy.attributes import __version__


extras_require = {
    'development': [
        'psycopg2-binary',
    ],
    "docs": [],
    'release': [
        'twine',
    ],
    'testing': [
        'tox',
        'watchdog',
    ],
}


install_requires = [
    'django>=2.0',
    'django-prettyjson',
    'djangorestframework>=3.10',
    'jsonschema',

    'huscy.pseudonyms',
]


setup(
    name='huscy.attributes',
    version=__version__,
    license='AGPLv3+',

    author='Alexander Tyapkov, Mathias Goldau, Stefan Bunde',
    author_email='tyapkov@gmail.com, goldau@cbs.mpg.de, stefanbunde+git@posteo.de',

    packages=find_namespace_packages(),

    install_requires=install_requires,

    extras_require=extras_require,

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
    ],
)
